@@ -1,6 +1,6 @@
 /* pkl.c - Poke compiler.  */
 
-/* Copyright (C) 2019, 2020, 2021 Jose E. Marchesi */
+/* Copyright (C) 2019, 2020, 2021, 2022 Jose E. Marchesi */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -72,8 +72,28 @@ struct pkl_compiler
   pkl_alien_token_handler_fn alien_token_fn;
 };
 
+
+static int
+pkl_load_rt (pkl_compiler compiler, char *poke_rt_pk)
+{
+  if (!pkl_execute_file (compiler, poke_rt_pk, NULL))
+    {
+      free (poke_rt_pk);
+
+      pk_term_class ("error");
+      pk_puts ("internal error: ");
+      pk_term_end_class ("error");
+      pk_puts ("compiler failed to bootstrap itself\n");
+
+      pkl_free (compiler);
+      return 0;
+    }
+  free (poke_rt_pk);
+  return 1;
+}
+
 pkl_compiler
-pkl_new (pvm vm, const char *rt_path)
+pkl_new (pvm vm, const char *rt_path, uint32_t flags)
 {
   pkl_compiler compiler
     = calloc (1, sizeof (struct pkl_compiler));
@@ -103,38 +123,35 @@ pkl_new (pvm vm, const char *rt_path)
     if (!poke_rt_pk)
       goto out_of_memory;
 
-    if (!pkl_execute_file (compiler, poke_rt_pk, NULL))
-      {
-        free (poke_rt_pk);
-
-        pk_term_class ("error");
-        pk_puts ("internal error: ");
-        pk_term_end_class ("error");
-        pk_puts ("compiler failed to bootstrap itself\n");
-
-        pkl_free (compiler);
-        return NULL;
-      }
-    free (poke_rt_pk);
+    if (!pkl_load_rt (compiler, poke_rt_pk))
+      return NULL;
 
     compiler->bootstrapped = 1;
   }
 
-  /* Load the standard library.  */
-  {
-    char *poke_std_pk = pk_str_concat (rt_path, "/std.pk", NULL);
-    if (!poke_std_pk)
-      goto out_of_memory;
+#define LOAD_STD(NAME)                                          \
+  do                                                            \
+    {                                                           \
+      char *path = pk_str_concat (rt_path, "/" NAME, NULL);     \
+      if (!path)                                                \
+        goto out_of_memory;                                     \
+                                                                \
+      if (!pkl_execute_file (compiler, path, NULL))             \
+        {                                                       \
+          free (path);                                          \
+          pkl_free (compiler);                                  \
+          return NULL;                                          \
+        }                                                       \
+                                                                \
+      free (path);                                              \
+    }                                                           \
+  while (0)
 
-    if (!pkl_execute_file (compiler, poke_std_pk, NULL))
-      {
-        free (poke_std_pk);
-        pkl_free (compiler);
-        return NULL;
-      }
-
-    free (poke_std_pk);
-  }
+  /* Load the standard library.  Note that the standard types may not
+     be loaded, depending on how the compiler is configured.  */
+  LOAD_STD ("std.pk");
+  if (!(flags & PKL_F_NOSTDTYPES))
+    LOAD_STD ("std-types.pk");
 
   return compiler;
 
@@ -164,7 +181,8 @@ pkl_free (pkl_compiler compiler)
 
 static pvm_program
 rest_of_compilation (pkl_compiler compiler,
-                     pkl_ast ast)
+                     pkl_ast ast,
+                     pkl_env env)
 {
   struct pkl_gen_payload gen_payload;
 
@@ -239,7 +257,7 @@ rest_of_compilation (pkl_compiler compiler,
   pkl_trans_init_payload (&trans2_payload);
   pkl_trans_init_payload (&trans3_payload);
   pkl_trans_init_payload (&trans4_payload);
-  pkl_gen_init_payload (&gen_payload, compiler);
+  pkl_gen_init_payload (&gen_payload, compiler, env);
 
   if (!pkl_do_pass (compiler, ast,
                     frontend_phases, frontend_payloads, PKL_PASS_F_TYPES, 1))
@@ -281,7 +299,8 @@ rest_of_compilation (pkl_compiler compiler,
 
 int
 pkl_execute_buffer (pkl_compiler compiler,
-                    const char *buffer, const char **end)
+                    const char *buffer, const char **end,
+                    pvm_val *exit_exception)
 {
   pkl_ast ast = NULL;
   pvm_program program;
@@ -300,9 +319,9 @@ pkl_execute_buffer (pkl_compiler compiler,
     goto error;
   else if (ret == 2)
     /* Memory exhaustion.  */
-    printf (_("out of memory\n"));
+    goto error;
 
-  program = rest_of_compilation (compiler, ast);
+  program = rest_of_compilation (compiler, ast, env);
   if (program == NULL)
     goto error;
 
@@ -312,10 +331,7 @@ pkl_execute_buffer (pkl_compiler compiler,
   /* Execute the program in the poke vm.  */
   {
     pvm_val val;
-
-    if (pvm_run (compiler->vm, program, &val) != PVM_EXIT_OK)
-      goto error;
-
+    pvm_run (compiler->vm, program, &val, exit_exception);
     /* Discard the value.  */
   }
 
@@ -332,7 +348,7 @@ pkl_execute_buffer (pkl_compiler compiler,
 int
 pkl_execute_statement (pkl_compiler compiler,
                        const char *buffer, const char **end,
-                       pvm_val *val)
+                       pvm_val *val, pvm_val *exit_exception)
 {
   pkl_ast ast = NULL;
   pvm_program program;
@@ -351,17 +367,16 @@ pkl_execute_statement (pkl_compiler compiler,
     goto error;
   else if (ret == 2)
     /* Memory exhaustion.  */
-    printf (_("out of memory\n"));
+    goto error;
 
-  program = rest_of_compilation (compiler, ast);
+  program = rest_of_compilation (compiler, ast, env);
   if (program == NULL)
     goto error;
 
   pvm_program_make_executable (program);
 
   /* Execute the routine in the poke vm.  */
-  if (pvm_run (compiler->vm, program, val) != PVM_EXIT_OK)
-    goto error;
+  pvm_run (compiler->vm, program, val, exit_exception);
 
   pvm_destroy_program (program);
   pkl_env_free (compiler->env);
@@ -394,9 +409,9 @@ pkl_compile_expression (pkl_compiler compiler,
      goto error;
    else if (ret == 2)
      /* Memory exhaustion.  */
-     printf (_("out of memory\n"));
+     goto error;
 
-   program = rest_of_compilation (compiler, ast);
+   program = rest_of_compilation (compiler, ast, env);
    if (program == NULL)
      goto error;
 
@@ -414,7 +429,7 @@ pkl_compile_expression (pkl_compiler compiler,
 int
 pkl_execute_expression (pkl_compiler compiler,
                         const char *buffer, const char **end,
-                        pvm_val *val)
+                        pvm_val *val, pvm_val *exit_exception)
 {
   pkl_ast ast = NULL;
   pvm_program program;
@@ -433,17 +448,16 @@ pkl_execute_expression (pkl_compiler compiler,
     goto error;
   else if (ret == 2)
     /* Memory exhaustion.  */
-    printf (_("out of memory\n"));
+    goto error;
 
-  program = rest_of_compilation (compiler, ast);
+  program = rest_of_compilation (compiler, ast, env);
   if (program == NULL)
     goto error;
 
   pvm_program_make_executable (program);
 
   /* Execute the routine in the poke vm.  */
-  if (pvm_run (compiler->vm, program, val) != PVM_EXIT_OK)
-    goto error;
+  pvm_run (compiler->vm, program, val, exit_exception);
 
   pvm_destroy_program (program);
   pkl_env_free (compiler->env);
@@ -457,7 +471,7 @@ pkl_execute_expression (pkl_compiler compiler,
 
 int
 pkl_execute_file (pkl_compiler compiler, const char *fname,
-                  int *exit_status)
+                  pvm_val *exit_exception)
 {
   int ret;
   pkl_ast ast = NULL;
@@ -480,12 +494,10 @@ pkl_execute_file (pkl_compiler compiler, const char *fname,
     /* Parse error.  */
     goto error;
   else if (ret == 2)
-    {
-      /* Memory exhaustion.  */
-      printf (_("out of memory\n"));
-    }
+    /* Memory exhaustion.  */
+    goto error;
 
-  program = rest_of_compilation (compiler, ast);
+  program = rest_of_compilation (compiler, ast, env);
   if (program == NULL)
     goto error;
 
@@ -495,10 +507,8 @@ pkl_execute_file (pkl_compiler compiler, const char *fname,
   /* Execute the program in the poke vm.  */
   {
     pvm_val val;
-    int status = pvm_run (compiler->vm, program, &val);
 
-    if (exit_status)
-      *exit_status = status;
+    pvm_run (compiler->vm, program, &val, exit_exception);
   }
 
   pvm_destroy_program (program);
@@ -589,22 +599,21 @@ pkl_set_alien_token_fn (pkl_compiler compiler,
 
 pvm_program
 pkl_compile_call (pkl_compiler compiler, pvm_val cls, pvm_val *ret,
-                  va_list ap)
+                  int narg, va_list ap)
 {
   pvm_program program;
   pkl_asm pasm;
   pvm_val arg;
+  int i;
 
   pasm = pkl_asm_new (NULL /* ast */, compiler, 1 /* prologue */);
 
   /* Push the arguments for the function.  */
-  do
+  for (i = 0; i < narg; ++i)
     {
       arg = va_arg (ap, pvm_val);
-      if (arg != PVM_NULL)
-        pkl_asm_insn (pasm, PKL_INSN_PUSH, arg);
+      pkl_asm_insn (pasm, PKL_INSN_PUSH, arg);
     }
-  while (arg != PVM_NULL);
 
   /* Call the closure.  */
   pkl_asm_insn (pasm, PKL_INSN_PUSH, cls);
