@@ -1,6 +1,6 @@
 /* pvm-val.c - Values for the PVM.  */
 
-/* Copyright (C) 2019, 2020, 2021 Jose E. Marchesi */
+/* Copyright (C) 2019, 2020, 2021, 2022 Jose E. Marchesi */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -84,6 +84,31 @@ pvm_make_ulong (uint64_t value, int size)
   return PVM_MAKE_LONG_ULONG (value, size, PVM_VAL_TAG_ULONG);
 }
 
+
+pvm_val
+pvm_make_signed_integral (int64_t value, int size)
+{
+  if (size > 64)
+    return PK_NULL;
+
+  if (size <= 32)
+    return PVM_MAKE_INT ((int32_t) value, size);
+  else
+    return PVM_MAKE_LONG (value, size);
+}
+
+pvm_val
+pvm_make_unsigned_integral (uint64_t value, int size)
+{
+  if (size > 64)
+    return PK_NULL;
+
+  if (size <= 32)
+    return PVM_MAKE_UINT ((uint32_t) value, size);
+  else
+    return PVM_MAKE_ULONG (value, size);
+}
+
 static pvm_val_box
 pvm_make_box (uint8_t tag)
 {
@@ -1007,11 +1032,12 @@ print_unit_name (uint64_t unit)
 }
 
 #define PVM_PRINT_VAL_1(...)                    \
-  pvm_print_val_1 (vm, depth, mode, base, indent, acutoff, flags, __VA_ARGS__)
+  pvm_print_val_1 (vm, depth, mode, base, indent, acutoff, flags, \
+                   exit_exception, __VA_ARGS__)
 
 static void
 pvm_print_val_1 (pvm vm, int depth, int mode, int base, int indent,
-                 int acutoff, uint32_t flags,
+                 int acutoff, uint32_t flags, pvm_val *exit_exception,
                  pvm_val val, int ndepth)
 {
   const char *long64_fmt, *long_fmt;
@@ -1061,9 +1087,9 @@ pvm_print_val_1 (pvm vm, int depth, int mode, int base, int indent,
       break;
     case 16:
       long64_fmt = "0x%" PRIx64 "L";
-      long_fmt = "(int<%d>) %" PRIx64;
+      long_fmt = "(int<%d>) 0x%" PRIx64;
       ulong64_fmt = "0x%" PRIx64 "UL";
-      ulong_fmt = "(uint<%d>) %" PRIx64;
+      ulong_fmt = "(uint<%d>) 0x%" PRIx64;
       int32_fmt = "0x%" PRIx32;
       int16_fmt = "0x%" PRIx32 "H";
       int8_fmt = "0x%" PRIx32 "B";
@@ -1341,7 +1367,7 @@ pvm_print_val_1 (pvm vm, int depth, int mode, int base, int indent,
          then use it, unless the PVM is configured to not do so.  */
       if (pprint)
         {
-          if (pvm_call_pretty_printer (vm, val))
+          if (pvm_call_pretty_printer (vm, val, exit_exception))
             return;
         }
 
@@ -1543,14 +1569,17 @@ pvm_print_val_1 (pvm vm, int depth, int mode, int base, int indent,
 #undef PVM_PRINT_VAL_1
 
 void
-pvm_print_val (pvm vm, pvm_val val)
+pvm_print_val (pvm vm, pvm_val val, pvm_val *exit_exception)
 {
+  if (exit_exception)
+    *exit_exception = PVM_NULL;
   pvm_print_val_1 (vm,
                    pvm_odepth (vm), pvm_omode (vm),
                    pvm_obase (vm), pvm_oindent (vm),
                    pvm_oacutoff (vm),
                    (pvm_omaps (vm) << (PVM_PRINT_F_MAPS - 1)
                     | (pvm_pretty_print (vm) << (PVM_PRINT_F_PPRINT - 1))),
+                   exit_exception,
                    val,
                    0 /* ndepth */);
 }
@@ -1559,11 +1588,15 @@ void
 pvm_print_val_with_params (pvm vm, pvm_val val,
                            int depth,int mode, int base,
                            int indent, int acutoff,
-                           uint32_t flags)
+                           uint32_t flags, pvm_val *exit_exception)
 {
+  if (exit_exception)
+    *exit_exception = PVM_NULL;
+
   pvm_print_val_1 (vm,
                    depth, mode, base, indent, acutoff,
                    flags,
+                   exit_exception,
                    val,
                    0 /* ndepth */);
 }
@@ -1594,6 +1627,8 @@ pvm_typeof (pvm_val val)
     type = PVM_VAL_ARR_TYPE (val);
   else if (PVM_IS_SCT (val))
     type = PVM_VAL_SCT_TYPE (val);
+  else if (PVM_IS_TYP (val))
+    type = val;
   else
     assert (0);
 
@@ -1672,14 +1707,14 @@ pvm_print_string (pvm_val string)
    corresponding to the struct VAL.  */
 
 int
-pvm_call_pretty_printer (pvm vm, pvm_val val)
+pvm_call_pretty_printer (pvm vm, pvm_val val, pvm_val *exit_exception)
 {
   pvm_val cls = pvm_get_struct_method (val, "_print");
 
   if (cls == PVM_NULL)
     return 0;
 
-  pvm_call_closure (vm, cls, val, PVM_NULL);
+  pvm_call_closure (vm, cls, exit_exception, val, PVM_NULL);
   return 1;
 }
 
@@ -1687,14 +1722,17 @@ pvm_call_pretty_printer (pvm vm, pvm_val val)
    definition of the struct Exception in pkl-rt.pk.  */
 
 pvm_val
-pvm_make_exception (int code, char *message, int exit_status)
+pvm_make_exception (int code, const char *name, int exit_status,
+                    const char *location, const char *msg)
 {
-  pvm_val nfields = pvm_make_ulong (3, 64);
+  pvm_val nfields = pvm_make_ulong (5, 64);
   pvm_val nmethods = pvm_make_ulong (0, 64);
   pvm_val struct_name = pvm_make_string ("Exception");
   pvm_val code_name = pvm_make_string ("code");
-  pvm_val msg_name = pvm_make_string ("msg");
+  pvm_val name_name = pvm_make_string ("name");
   pvm_val exit_status_name = pvm_make_string ("exit_status");
+  pvm_val location_name = pvm_make_string ("location");
+  pvm_val msg_name = pvm_make_string ("msg");
   pvm_val *field_names, *field_types, type;
   pvm_val exception;
 
@@ -1704,9 +1742,19 @@ pvm_make_exception (int code, char *message, int exit_status)
   field_types[0] = pvm_make_integral_type (pvm_make_ulong (32, 64),
                                            PVM_MAKE_INT (1, 32));
 
-  field_names[1] = msg_name;
+  field_names[1] = name_name;
   field_types[1] = pvm_make_string_type ();
 
+  field_names[2] = exit_status_name;
+  field_types[2] = pvm_make_integral_type (pvm_make_ulong (32, 64),
+                                           PVM_MAKE_INT (1, 32));
+
+  field_names[3] = location_name;
+  field_types[3] = pvm_make_string_type ();
+
+  field_names[4] = msg_name;
+  field_types[4] = pvm_make_string_type ();
+
   type = pvm_make_struct_type (nfields, struct_name,
                                field_names, field_types);
 
@@ -1716,14 +1764,22 @@ pvm_make_exception (int code, char *message, int exit_status)
   PVM_VAL_SCT_FIELD_VALUE (exception, 0)
     = PVM_MAKE_INT (code, 32);
 
-  PVM_VAL_SCT_FIELD_NAME (exception, 1) = msg_name;
+  PVM_VAL_SCT_FIELD_NAME (exception, 1) = name_name;
   PVM_VAL_SCT_FIELD_VALUE (exception, 1)
-    = pvm_make_string (message);
+    = pvm_make_string (name);
 
   PVM_VAL_SCT_FIELD_NAME (exception, 2) = exit_status_name;
   PVM_VAL_SCT_FIELD_VALUE (exception, 2)
     = PVM_MAKE_INT (exit_status, 32);
 
+  PVM_VAL_SCT_FIELD_NAME (exception, 3) = location_name;
+  PVM_VAL_SCT_FIELD_VALUE (exception, 3)
+    = pvm_make_string (location == NULL ? "" : location);
+
+  PVM_VAL_SCT_FIELD_NAME (exception, 4) = msg_name;
+  PVM_VAL_SCT_FIELD_VALUE (exception, 4)
+    = pvm_make_string (msg == NULL ? "" : msg);
+
   return exception;
 }
 
