@@ -1,6 +1,6 @@
 /* pvm.h - Poke Virtual Machine.  */
 
-/* Copyright (C) 2019, 2020, 2021 Jose E. Marchesi */
+/* Copyright (C) 2019, 2020, 2021, 2022 Jose E. Marchesi */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -210,6 +210,15 @@ pvm_val pvm_make_long (int64_t value, int size);
 
 pvm_val pvm_make_ulong (uint64_t value, int size);
 
+/* Make signed and unsigned integral values.
+
+   SIZE is measured in bits and should be in the range 1 to 64.  If an
+   invalid size is provided these functions return PVM_NULL.  */
+
+pvm_val pvm_make_signed_integral (int64_t value, int size);
+
+pvm_val pvm_make_unsigned_integral (uint64_t value, int size);
+
 /* Make a string PVM value.  */
 
 pvm_val pvm_make_string (const char *value);
@@ -364,11 +373,14 @@ void pvm_val_ureloc (pvm_val val);
    map-able then this is a no-operation.  */
 void pvm_val_unmap (pvm_val val);
 
-/* Return a PVM value for an exception with the given CODE, MESSAGE
-   and EXIT_STATUS.  */
+/* Return a PVM value for an exception with the given CODE, NAME,
+   EXIT_STATUS, LOCATION and MSG.
 
-pvm_val pvm_make_exception (int code, char *message, int exit_status);
+   Both LOCATION and MSG are optional.  If NULL is passed for them
+   then a suitable default value is used.  */
 
+pvm_val pvm_make_exception (int code, const char *name, int exit_status,
+                            const char *location, const char *message);
 
 /* **************** The Run-Time Environment ****************  */
 
@@ -476,83 +488,83 @@ enum pvm_exit_code
     PVM_EXIT_ERROR = 1
   };
 
-/* Exceptions.  These should be in sync with the exception code
-   variables, and the exception messages, declared in pkl-rt.pkl */
+/* Exceptions.  These should be in sync with the exception code macros
+   in libpoke.h, and variables and exception names, declared in pkl-rt.pk */
 
 #define PVM_E_GENERIC       0
-#define PVM_E_GENERIC_MSG "generic"
+#define PVM_E_GENERIC_NAME "generic"
 #define PVM_E_GENERIC_ESTATUS 1
 
 #define PVM_E_DIV_BY_ZERO   1
-#define PVM_E_DIV_BY_ZERO_MSG "division by zero"
+#define PVM_E_DIV_BY_ZERO_NAME "division by zero"
 #define PVM_E_DIV_BY_ZERO_ESTATUS 1
 
 #define PVM_E_NO_IOS        2
-#define PVM_E_NO_IOS_MSG "no IOS"
+#define PVM_E_NO_IOS_NAME "no IOS"
 #define PVM_E_NO_IOS_ESTATUS 1
 
 #define PVM_E_NO_RETURN     3
-#define PVM_E_NO_RETURN_MSG "no return"
+#define PVM_E_NO_RETURN_NAME "no return"
 #define PVM_E_NO_RETURN_ESTATUS 1
 
 #define PVM_E_OUT_OF_BOUNDS 4
-#define PVM_E_OUT_OF_BOUNDS_MSG "out of bounds"
+#define PVM_E_OUT_OF_BOUNDS_NAME "out of bounds"
 #define PVM_E_OUT_OF_BOUNDS_ESTATUS 1
 
 #define PVM_E_MAP_BOUNDS    5
-#define PVM_E_MAP_BOUNDS_MSG "out of map bounds"
+#define PVM_E_MAP_BOUNDS_NAME "out of map bounds"
 #define PVM_E_MAP_BOUNDS_ESTATUS 1
 
 #define PVM_E_EOF           6
-#define PVM_E_EOF_MSG "EOF"
+#define PVM_E_EOF_NAME "EOF"
 #define PVM_E_EOF_ESTATUS 1
 
 #define PVM_E_MAP           7
-#define PVM_E_MAP_MSG "no map"
+#define PVM_E_MAP_NAME "no map"
 #define PVM_E_MAP_ESTATUS 1
 
 #define PVM_E_CONV          8
-#define PVM_E_CONV_MSG "conversion error"
+#define PVM_E_CONV_NAME "conversion error"
 #define PVM_E_CONV_ESTATUS 1
 
 #define PVM_E_ELEM          9
-#define PVM_E_ELEM_MSG "invalid element"
+#define PVM_E_ELEM_NAME "invalid element"
 #define PVM_E_ELEM_ESTATUS 1
 
 #define PVM_E_CONSTRAINT   10
-#define PVM_E_CONSTRAINT_MSG "constraint violation"
+#define PVM_E_CONSTRAINT_NAME "constraint violation"
 #define PVM_E_CONSTRAINT_ESTATUS 1
 
 #define PVM_E_IO           11
-#define PVM_E_IO_MSG "generic IO"
+#define PVM_E_IO_NAME "generic IO"
 #define PVM_E_IO_ESTATUS 1
 
 #define PVM_E_SIGNAL       12
-#define PVM_E_SIGNAL_MSG ""
+#define PVM_E_SIGNAL_NAME ""
 #define PVM_E_SIGNAL_ESTATUS 1
 
 #define PVM_E_IOFLAGS      13
-#define PVM_E_IOFLAGS_MSG "invalid IO flags"
+#define PVM_E_IOFLAGS_NAME "invalid IO flags"
 #define PVM_E_IOFLAGS_ESTATUS 1
 
 #define PVM_E_INVAL        14
-#define PVM_E_INVAL_MSG "invalid argument"
+#define PVM_E_INVAL_NAME "invalid argument"
 #define PVM_E_INVAL_ESTATUS 1
 
 #define PVM_E_EXIT         15
-#define PVM_E_EXIT_MSG ""
+#define PVM_E_EXIT_NAME ""
 #define PVM_E_EXIT_ESTATUS 0
 
 #define PVM_E_ASSERT       16
-#define PVM_E_ASSERT_MSG "assertion failure"
+#define PVM_E_ASSERT_NAME "assertion failure"
 #define PVM_E_ASSERT_ESTATUS 1
 
 #define PVM_E_OVERFLOW     17
-#define PVM_E_OVERFLOW_MSG "overflow"
+#define PVM_E_OVERFLOW_NAME "overflow"
 #define PVM_E_OVERFLOW_ESTATUS 1
 
 #define PVM_E_PERM         18
-#define PVM_E_PERM_MSG     "wrong permissions"
+#define PVM_E_PERM_NAME     "wrong permissions"
 #define PVM_E_PERM_ESTATUS 1
 
 typedef struct pvm *pvm;
@@ -583,12 +595,17 @@ void pvm_reset_profile (pvm pvm);
    If the execution of PROGRAM generates a result value, it is put in
    RES.
 
+   If not NULL, *EXIT_EXCEPTION is set to an exception value if the
+   execution of the program gets interrupted by an unhandled exception.
+   Otherwise *EXIT_EXCEPTION is set to PK_NULL.
+
    This function returns an exit code, indicating whether the
    execution was successful or not.  */
 
 enum pvm_exit_code pvm_run (pvm vm,
                             pvm_program program,
-                            pvm_val *res);
+                            pvm_val *res,
+                            pvm_val *exit_exception);
 
 /* Given a PVM and a closure value, call the closure.
 
@@ -596,7 +613,8 @@ enum pvm_exit_code pvm_run (pvm vm,
    arguments to the closure.  Note that this function doesn't allow
    pass nulls as arguments.  */
 
-void pvm_call_closure (pvm vm, pvm_val cls, ...);
+void pvm_call_closure (pvm vm, pvm_val cls, pvm_val *exit_exception,
+                       ...);
 
 /* Get/set the current byte endianness of a virtual machine.
 
@@ -698,7 +716,8 @@ void pvm_handle_signal (int signal_number);
 
 /* Call the pretty printer of the given value VAL.  */
 
-int pvm_call_pretty_printer (pvm vm, pvm_val val);
+int pvm_call_pretty_printer (pvm vm, pvm_val val,
+                             pvm_val *exit_exception);
 
 /* Print a PVM value.
 
@@ -732,11 +751,12 @@ int pvm_call_pretty_printer (pvm vm, pvm_val val);
 #define PVM_PRINT_F_MAPS   1
 #define PVM_PRINT_F_PPRINT 2
 
-void pvm_print_val (pvm vm, pvm_val val);
+void pvm_print_val (pvm vm, pvm_val val, pvm_val *exit_exception);
 
 void pvm_print_val_with_params (pvm vm, pvm_val val,
                                 int depth,int mode, int base,
                                 int indent, int acutoff,
-                                uint32_t flags);
+                                uint32_t flags,
+                                pvm_val *exit_exception);
 
 #endif /* ! PVM_H */
