@@ -1,6 +1,6 @@
 /* pk-cmd.c - Poke commands.  */
 
-/* Copyright (C) 2019, 2020, 2021 Jose E. Marchesi */
+/* Copyright (C) 2019, 2020, 2021, 2022 Jose E. Marchesi */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -35,10 +35,16 @@
 #include "pk-cmd.h"
 #include "pk-utils.h"
 
+/* This function is defined in pk-cmd-set.c and used below in
+   pk_cmd_init.  */
+extern void pk_cmd_set_init (void);
+
 /* Table of supported commands.  */
 
 extern const struct pk_cmd ios_cmd; /* pk-cmd-ios.c */
 extern const struct pk_cmd file_cmd; /* pk-cmd-ios.c  */
+extern const struct pk_cmd proc_cmd; /* pk-cmd-ios.c */
+extern const struct pk_cmd sub_cmd;  /* pk-cmd-sub.c */
 extern const struct pk_cmd mem_cmd; /* pk-cmd-ios.c */
 #ifdef HAVE_LIBNBD
 extern const struct pk_cmd nbd_cmd; /* pk-cmd-ios.c */
@@ -64,6 +70,8 @@ static const struct pk_cmd *dot_cmds[] =
   {
     &ios_cmd,
     &file_cmd,
+    &proc_cmd,
+    &sub_cmd,
     &exit_cmd,
     &quit_cmd,
     &version_cmd,
@@ -274,10 +282,11 @@ pk_cmd_exec_1 (const char *str, struct pk_trie *cmds_trie, char *prefix)
   const char *p;
   const struct pk_cmd *cmd;
   int argc = 0;
-  struct pk_cmd_arg argv[8];
+  struct pk_cmd_arg argv[9];
   uint64_t uflags;
   const char *a;
   int besilent = 0;
+  int run_default_handler_p = 0;
 
   /* Skip blanks, and return if the command is composed by only blank
      characters.  */
@@ -306,6 +315,12 @@ pk_cmd_exec_1 (const char *str, struct pk_trie *cmds_trie, char *prefix)
       pk_printf (_("%s: command not found.\n"), cmd_name);
       return 0;
     }
+  strncpy (cmd_name, cmd->name, MAX_CMD_NAME - 1);
+
+  /* argv[0] is the command name.  */
+  argv[argc].type = PK_CMD_ARG_STR;
+  argv[argc].val.str = xstrdup (cmd_name);
+  argc += 1;
 
   /* Process user flags.  */
   uflags = 0;
@@ -337,7 +352,10 @@ pk_cmd_exec_1 (const char *str, struct pk_trie *cmds_trie, char *prefix)
     {
       p = skip_blanks (p);
       if (*p == '\0')
-        GOTO_USAGE();
+        {
+          run_default_handler_p = 1;
+          GOTO_USAGE();
+        }
       return pk_cmd_exec_1 (p, *cmd->subtrie, cmd_name);
     }
 
@@ -536,6 +554,13 @@ pk_cmd_exec_1 (const char *str, struct pk_trie *cmds_trie, char *prefix)
 
   besilent = 1;
   usage:
+  if (!besilent && run_default_handler_p)
+    {
+      if (cmd->handler)
+        ret = (*cmd->handler) (argc, argv, uflags);
+      else
+        run_default_handler_p = 0;
+    }
   /* Free arguments occupying memory.  */
   for (int i = 0; i < argc; ++i)
     {
@@ -543,7 +568,7 @@ pk_cmd_exec_1 (const char *str, struct pk_trie *cmds_trie, char *prefix)
         free (argv[i].val.str);
     }
 
-  if (!besilent)
+  if (!besilent && !run_default_handler_p)
     pk_printf (_("Usage: %s\n"), cmd->usage);
 
   return ret;
@@ -562,7 +587,7 @@ extern struct pk_trie *vm_disas_trie; /* pk-cmd-vm.c */
 extern const struct pk_cmd *vm_profile_cmds[]; /* pk-cmd-vm.c */
 extern struct pk_trie *vm_profile_trie; /* pk-cmd-vm.c */
 
-extern const struct pk_cmd *set_cmds[]; /* pk-cmd-set.c */
+extern const struct pk_cmd **set_cmds; /* pk-cmd-set.c */
 extern struct pk_trie *set_trie; /* pk-cmd-set.c */
 
 extern const struct pk_cmd *map_cmds[]; /* pk-cmd-map.c */
@@ -594,6 +619,7 @@ pk_cmd_exec (const char *str)
       char *cmd_alloc = NULL;
       int what; /* 0 -> declaration, 1 -> statement */
       int retval = 1;
+      pk_val exit_exception;
 
       if (IS_COMMAND(ecmd, "fun"))
         what = 0;
@@ -617,7 +643,8 @@ pk_cmd_exec (const char *str)
       if (what == 0)
         {
           /* Declaration.  */
-          if (pk_compile_buffer (poke_compiler, ecmd, &end) != PK_OK)
+          if (pk_compile_buffer (poke_compiler, ecmd, &end,
+                                 &exit_exception) != PK_OK)
             {
               retval = 0;
               goto cleanup;
@@ -628,7 +655,8 @@ pk_cmd_exec (const char *str)
           /* Statement.  */
           pk_val val;
 
-          if (pk_compile_statement (poke_compiler, ecmd, &end, &val) != PK_OK)
+          if (pk_compile_statement (poke_compiler, ecmd, &end, &val,
+                                    &exit_exception) != PK_OK)
             {
               retval = 0;
               goto cleanup;
@@ -636,12 +664,29 @@ pk_cmd_exec (const char *str)
 
           if (val != PK_NULL)
             {
-              pk_print_val (poke_compiler, val);
-              pk_puts ("\n");
+              /* Note that printing the value may result in an
+                 exception, in case a pretty-printer is involved.  */
+              pk_print_val (poke_compiler, val, &exit_exception);
+              if (exit_exception == PK_NULL)
+                pk_puts ("\n");
             }
         }
       pk_set_lexical_cuckolding_p (poke_compiler, 0);
 
+      /* If the execution (or the printing of the value) ended due to
+         an unhandled exception (this may include E_exit) then call
+         the poke default handler.  */
+      if (exit_exception != PK_NULL)
+        {
+          pk_val default_handler
+            = pk_decl_val (poke_compiler, "pk_exception_handler");
+
+          assert (default_handler != PK_NULL);
+          if (pk_call (poke_compiler, default_handler, NULL,
+                       1, exit_exception) == PK_ERROR)
+            assert (0);
+        }
+
     cleanup:
       free (cmd_alloc);
       return retval;
@@ -724,10 +769,13 @@ pk_cmd_init (void)
   vm_trie = pk_trie_from_cmds (vm_cmds);
   vm_disas_trie = pk_trie_from_cmds (vm_disas_cmds);
   vm_profile_trie = pk_trie_from_cmds (vm_profile_cmds);
-  set_trie = pk_trie_from_cmds (set_cmds);
   map_trie = pk_trie_from_cmds (map_cmds);
   map_entry_trie = pk_trie_from_cmds (map_entry_cmds);
 
+  /* The set_cmds are built dynamically.  */
+  pk_cmd_set_init ();
+  set_trie = pk_trie_from_cmds (set_cmds);
+
   /* Compile commands written in Poke.  */
   if (!pk_load (poke_compiler, "pk-cmd"))
     pk_fatal ("unable to load the pk-cmd module");
