@@ -1,6 +1,6 @@
 /* poke.c - Interactive editor for binary files.  */
 
-/* Copyright (C) 2019, 2020, 2021 Jose E. Marchesi */
+/* Copyright (C) 2019, 2020, 2021, 2022 Jose E. Marchesi */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,8 +25,8 @@
 #include <string.h>
 #include <locale.h>
 #include <textstyle.h>
-#include "xalloc.h"
 #include <assert.h>
+#include <arpa/inet.h> /* For htonl */
 
 #ifdef HAVE_HSERVER
 #  include "pk-hserver.h"
@@ -103,27 +103,17 @@ char *poke_mapsdir;
 
 char *poke_docdir;
 
-/* The following global contains the name of the program to use to
-   display documentation.  Valid values are `info' and `less'.  It
-   defaults to `info'.  */
-
-char *poke_doc_viewer = NULL;
-
 /* The following global determines whether auto-maps shall be
-   acknowleged when loading files or not.  Defaults to `yes'.  */
+   acknowleged when loading files by default.  Defaults to 1, but can
+   be changed to 0 using the --no-auto-map command-line option.  */
 
-int poke_auto_map_p = 1;
+int poke_default_auto_map_p = 1;
 
 /* The following global determines whether the user specified that the
    hyperlinks server should not be activated, in the command line.  */
 
 int poke_no_hserver_arg = 0;
 
-/* The following global determines whether map information shall be
-   included in the REPL prompt.  Defaults to `yes'.  */
-
-int poke_prompt_maps_p = 1;
-
 /* This is used by commands to indicate to the REPL that it must
    exit.  */
 
@@ -152,6 +142,8 @@ enum
   SOURCE_ARG,
   COLOR_ARG,
   STYLE_ARG,
+  STYLE_DARK_ARG,
+  STYLE_BRIGHT_ARG,
   MI_ARG,
   NO_AUTO_MAP_ARG,
   NO_HSERVER_ARG
@@ -168,6 +160,8 @@ static const struct option long_options[] =
   {"no-init-file", no_argument, NULL, NO_INIT_FILE_ARG},
   {"color", required_argument, NULL, COLOR_ARG},
   {"style", required_argument, NULL, STYLE_ARG},
+  {"style-dark", no_argument, NULL, STYLE_DARK_ARG},
+  {"style-bright", no_argument, NULL, STYLE_BRIGHT_ARG},
   {"mi", no_argument, NULL, MI_ARG},
   {"no-auto-map", no_argument, NULL, NO_AUTO_MAP_ARG},
   {"no-hserver", no_argument, NULL, NO_HSERVER_ARG},
@@ -203,6 +197,8 @@ print_help (void)
   pk_puts (_("Styling text output:\n"));
   pk_puts (_("      --color=(yes|no|auto|html|test) emit styled output\n"));
   pk_puts (_("      --style=STYLE_FILE              style file to use when styling\n"));
+  pk_puts (_("      --style-dark                    use default style for dark backgrounds\n"));
+  pk_puts (_("      --style-bright                  use default style for bright backgrounds\n"));
 
   pk_puts ("\n");
   pk_puts (_("Machine interface:\n"));
@@ -316,6 +312,31 @@ static struct pk_term_if poke_term_if =
     .set_bgcolor_fn = pk_term_set_bgcolor,
   };
 
+const char *
+pk_var_string (const char *name)
+{
+  pk_val val = pk_decl_val (poke_compiler, name);
+
+  assert (val != PK_NULL);
+  return pk_string_str (val);
+}
+
+int
+pk_var_int (const char *name)
+{
+  pk_val val = pk_decl_val (poke_compiler, name);
+
+  assert (val != PK_NULL);
+  return pk_int_value (val);
+}
+
+void
+pk_set_var_int (const char *name, int value)
+{
+  pk_decl_set_val (poke_compiler, name,
+                   pk_make_int (value, 32));
+}
+
 static void
 set_script_args (int argc, char *argv[])
 {
@@ -373,7 +394,7 @@ parse_args_1 (int argc, char *argv[])
           poke_interactive_p = 0;
           return;
         case NO_AUTO_MAP_ARG:
-          poke_auto_map_p = 0;
+          poke_default_auto_map_p = 0;
           break;
         case NO_HSERVER_ARG:
           poke_no_hserver_arg = 1;
@@ -418,10 +439,24 @@ parse_args_2 (int argc, char *argv[])
           break;
         case 'l':
         case LOAD_ARG:
-          if (pk_compile_file (poke_compiler, optarg,
-                               NULL /* exit_status */) != PK_OK)
-            goto exit_success;
-          break;
+          {
+            pk_val exception;
+
+            if (pk_compile_file (poke_compiler, optarg, &exception) != PK_OK)
+              goto exit_success;
+            if (exception != PK_NULL)
+              {
+                pk_val default_handler
+                  = pk_decl_val (poke_compiler, "pk_exception_handler");
+
+                assert (default_handler != PK_NULL);
+                if (pk_call (poke_compiler, default_handler, NULL,
+                             1, exception) == PK_ERROR)
+                  assert (0);
+                goto exit_success;
+              }
+            break;
+          }
         case 'c':
         case CMD_ARG:
           {
@@ -439,14 +474,19 @@ parse_args_2 (int argc, char *argv[])
           }
         case 'L':
           {
-            int exit_status;
+            pk_val exception;
+            int exit_status = 0;
 
             /* Build argv in the compiler, with the rest of the
                command-line arguments.  Then execute the script and
                return.  */
             set_script_args (argc, argv);
-            if (pk_compile_file (poke_compiler, optarg, &exit_status) != PK_OK)
+            if (pk_compile_file (poke_compiler, optarg, &exception) != PK_OK)
               goto exit_failure;
+            if (exception != PK_NULL)
+              exit_status
+                = pk_int_value (pk_struct_ref_field_value (exception,
+                                                           "exit_status"));
 
             finalize ();
             exit (exit_status);
@@ -463,6 +503,8 @@ parse_args_2 (int argc, char *argv[])
              here.   */
         case COLOR_ARG:
         case STYLE_ARG:
+        case STYLE_DARK_ARG:
+        case STYLE_BRIGHT_ARG:
           break;
         default:
           goto exit_failure;
@@ -553,10 +595,61 @@ initialize (int argc, char *argv[])
                  pk_make_int (poke_interactive_p, 32)) == PK_ERROR)
     pk_fatal ("defining poke_interactive_p");
 
+  /* The Poke hyperlinks facilities must be loaded before poke.pk and
+     the cmd subsystem.  This is done even if the hserver is
+     disabled.  */
+  if (!pk_load (poke_compiler, "pk-hserver"))
+    pk_fatal ("unable to load the pk-hserver module");
+
   /* Load poke.pk  */
   if (!pk_load (poke_compiler, "poke"))
     pk_fatal ("unable to load the poke module");
 
+  /* Set the values of a few global variables defined in poke.pk.  */
+  {
+    uint32_t canary = 0x11223344;
+    pk_val endian_big = pk_decl_val (poke_compiler, "ENDIAN_BIG");
+    pk_val endian_little = pk_decl_val (poke_compiler, "ENDIAN_LITTLE");
+    pk_val host_endian;
+    pk_val network_endian;
+
+#ifdef WORDS_BIGENDIAN
+    host_endian = endian_big;
+#else
+    host_endian = endian_little;
+#endif
+
+    if (canary == htonl (canary))
+      {
+#ifdef WORDS_BIGENDIAN
+        network_endian = endian_big;
+#else
+        network_endian = endian_little;
+#endif
+      }
+    else
+      {
+#ifdef WORDS_BIGENDIAN
+        network_endian = endian_little;
+#else
+        network_endian = endian_big;
+#endif
+      }
+
+    pk_decl_set_val (poke_compiler, "pk_host_endian", host_endian);
+    pk_decl_set_val (poke_compiler, "pk_network_endian", network_endian);
+
+    pk_decl_set_val (poke_compiler, "pk_auto_map_p",
+                     pk_make_int (poke_default_auto_map_p, 32));
+
+    pk_decl_set_val (poke_compiler, "pk_have_libnbd_p",
+#if defined HAVE_LIBNBD
+                     pk_make_int (1, 32)
+#else
+                     pk_make_int (0, 32)
+#endif
+                     );
+  }
 
   /* Initialize the global map.  */
   pk_map_init ();
@@ -577,10 +670,11 @@ initialize (int argc, char *argv[])
       pk_hserver_init ();
       pk_hserver_start ();
     }
-#endif
 
-  /* Initialize the documentation viewer.  */
-  poke_doc_viewer = xstrdup ("info");
+  /* Set the value of the Poke variable pk_hserver_p.  */
+  pk_decl_set_val (poke_compiler, "pk_hserver_p",
+                   pk_make_int (poke_hserver_p, 32));
+#endif
 }
 
 static void
