@@ -1,6 +1,6 @@
 /* poke.h - Interactive editor for binary files.  */
 
-/* Copyright (C) 2019, 2020, 2021 Jose E. Marchesi */
+/* Copyright (C) 2019, 2020, 2021, 2022 Jose E. Marchesi */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -42,9 +42,10 @@ extern char *poke_picklesdir;
 extern char *poke_mapsdir;
 extern char *poke_docdir;
 extern char *poke_cmdsdir;
-extern char *poke_doc_viewer;
-extern int poke_auto_map_p;
-extern int poke_prompt_maps_p;
+
+int pk_var_int (const char *name);
+void pk_set_var_int (const char *name, int value);
+const char *pk_var_string (const char *name);
 
 void pk_print_version (int hand_p);
 void pk_fatal (const char *errmsg) __attribute__ ((noreturn));
